"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var path_1 = __importDefault(require("path"));
var createControllersText_1 = __importDefault(require("./createControllersText"));
var checkRequisites_1 = __importDefault(require("./checkRequisites"));
var genHandlerText = function (isAsync) { return "\nconst " + (isAsync ? 'asyncM' : 'm') + "ethodToHandler = (\n  methodCallback: ServerMethods<any, any>[LowerHttpMethod]\n): RequestHandler => " + (isAsync ? 'async ' : '') + "(req, res, next) => {\n  try {\n    const data = " + (isAsync ? 'await ' : '') + "methodCallback(req as any) as any\n\n    if (data.headers) {\n      for (const key in data.headers) {\n        res.setHeader(key, data.headers[key])\n      }\n    }\n\n    res.status(data.status).send(data.body)\n  } catch (e) {\n    next(e)\n  }\n}\n"; };
var genHandlerWithSchemaText = function (isAsync) { return "\nconst " + (isAsync ? 'asyncM' : 'm') + "ethodToHandlerWithSchema = (\n  methodCallback: ServerMethods<any, any>[LowerHttpMethod],\n  schema: { [K in HttpStatusOk]?: Schema }\n): RequestHandler => {\n  const stringifySet = Object.entries(schema).reduce(\n    (prev, [key, val]) => ({ ...prev, [key]: fastJson(val!) }),\n    {} as Record<HttpStatusOk, ReturnType<typeof fastJson> | undefined>\n  )\n\n  return " + (isAsync ? 'async ' : '') + "(req, res, next) => {\n    try {\n      const data = " + (isAsync ? 'await ' : '') + "methodCallback(req as any) as any\n      const stringify = stringifySet[data.status as HttpStatusOk]\n\n      if (stringify) {\n        res.set('content-type', 'application/json; charset=utf-8')\n\n        if (data.headers) {\n          for (const key in data.headers) {\n            res.setHeader(key, data.headers[key])\n          }\n        }\n\n        res.status(data.status).send(stringify(data.body))\n      } else {\n        if (data.headers) {\n          for (const key in data.headers) {\n            res.setHeader(key, data.headers[key])\n          }\n        }\n\n        res.status(data.status).send(data.body)\n      }\n    } catch (e) {\n      next(e)\n    }\n  }\n}\n"; };
exports.default = (function (input, project) {
    var _a = createControllersText_1.default(input + "/api", project !== null && project !== void 0 ? project : input), imports = _a.imports, consts = _a.consts, controllers = _a.controllers;
    var hasNumberTypeQuery = controllers.includes('parseNumberTypeQueryParams(');
    var hasBooleanTypeQuery = controllers.includes('parseBooleanTypeQueryParams(');
    var hasOptionalQuery = controllers.includes('  callParserIfExistsQuery(');
    var hasJSONBody = controllers.includes('  parseJSONBoby,');
    var hasTypedParams = controllers.includes('  createTypedParamsHandler(');
    var hasValidator = controllers.includes('  validateOrReject(');
    var hasMulter = controllers.includes('  uploader,');
    var hasMethodToHandler = controllers.includes(' methodToHandler(');
    var hasAsyncMethodToHandler = controllers.includes(' asyncMethodToHandler(');
    var hasMethodToHandlerWithSchema = controllers.includes(' methodToHandlerWithSchema(');
    var hasAsyncMethodToHandlerWithSchema = controllers.includes(' asyncMethodToHandlerWithSchema(');
    checkRequisites_1.default({ hasValidator: hasValidator });
    var headIpmorts = [];
    if (hasValidator) {
        headIpmorts.push("import 'reflect-metadata'", "import type { ClassTransformOptions } from 'class-transformer'", "import { plainToInstance } from 'class-transformer'", "import type { ValidatorOptions } from 'class-validator'", "import { validateOrReject } from 'class-validator'");
    }
    if (hasMulter) {
        headIpmorts.push("import path from 'path'");
    }
    headIpmorts.push("import type { Express, RequestHandler" + (hasValidator ? ', Request' : '') + " } from 'express'");
    if (hasJSONBody) {
        headIpmorts.push("import express from 'express'");
    }
    if (hasMulter) {
        headIpmorts.push("import type { Options } from 'multer'");
        headIpmorts.push("import multer from 'multer'");
    }
    if (hasMethodToHandlerWithSchema || hasAsyncMethodToHandlerWithSchema) {
        headIpmorts.push("import type { Schema } from 'fast-json-stringify'");
        headIpmorts.push("import fastJson from 'fast-json-stringify'");
    }
    if (hasValidator) {
        headIpmorts.push("import * as Validators from './validators'");
    }
    if (hasMulter) {
        headIpmorts.push("import type { ReadStream } from 'fs'");
    }
    headIpmorts.push("import type { LowerHttpMethod, AspidaMethods, HttpStatusOk, AspidaMethodParams } from 'aspida'");
    return {
        text: headIpmorts.join('\n') + "\n" + imports + "\n\nexport type FrourioOptions = {\n  basePath?: string\n" + (hasValidator ? '  transformer?: ClassTransformOptions\n  validator?: ValidatorOptions\n' : '') + (hasMulter
            ? "  multer?: Options\n}\n\nexport type MulterFile = Express.Multer.File"
            : '}') + "\n\ntype HttpStatusNoOk = 301 | 302 | 400 | 401 | 402 | 403 | 404 | 405 | 406 | 409 | 500 | 501 | 502 | 503 | 504 | 505\n\ntype PartiallyPartial<T, K extends keyof T> = Omit<T, K> & Partial<Pick<T, K>>\n\ntype BaseResponse<T, U, V> = {\n  status: V extends number ? V : HttpStatusOk\n  body: T\n  headers: U\n}\n\ntype ServerResponse<K extends AspidaMethodParams> =\n  | (K extends { resBody: K['resBody']; resHeaders: K['resHeaders'] }\n  ? BaseResponse<K['resBody'], K['resHeaders'], K['status']>\n  : K extends { resBody: K['resBody'] }\n  ? PartiallyPartial<BaseResponse<K['resBody'], K['resHeaders'], K['status']>, 'headers'>\n  : K extends { resHeaders: K['resHeaders'] }\n  ? PartiallyPartial<BaseResponse<K['resBody'], K['resHeaders'], K['status']>, 'body'>\n  : PartiallyPartial<\n      BaseResponse<K['resBody'], K['resHeaders'], K['status']>,\n      'body' | 'headers'\n    >)\n  | PartiallyPartial<BaseResponse<any, any, HttpStatusNoOk>, 'body' | 'headers'>\n" + (hasMulter
            ? "\ntype BlobToFile<T extends AspidaMethodParams> = T['reqFormat'] extends FormData\n  ? {\n      [P in keyof T['reqBody']]: Required<T['reqBody']>[P] extends Blob | ReadStream\n        ? MulterFile\n        : Required<T['reqBody']>[P] extends (Blob | ReadStream)[]\n        ? MulterFile[]\n        : T['reqBody'][P]\n    }\n  : T['reqBody']\n"
            : '') + "\ntype RequestParams<T extends AspidaMethodParams> = Pick<{\n  query: T['query']\n  body: " + (hasMulter ? 'BlobToFile<T>' : "T['reqBody']") + "\n  headers: T['reqHeaders']\n}, {\n  query: Required<T>['query'] extends {} | null ? 'query' : never\n  body: Required<T>['reqBody'] extends {} | null ? 'body' : never\n  headers: Required<T>['reqHeaders'] extends {} | null ? 'headers' : never\n}['query' | 'body' | 'headers']>\n\nexport type ServerMethods<T extends AspidaMethods, U extends Record<string, any> = {}> = {\n  [K in keyof T]: (\n    req: RequestParams<T[K]> & U\n  ) => ServerResponse<T[K]> | Promise<ServerResponse<T[K]>>\n}\n" + (hasNumberTypeQuery
            ? "\nconst parseNumberTypeQueryParams = (numberTypeParams: [string, boolean, boolean][]): RequestHandler => ({ query }, res, next) => {\n  for (const [key, isOptional, isArray] of numberTypeParams) {\n    const param = query[key]\n\n    if (isArray) {\n      if (!isOptional && param === undefined) {\n        query[key] = []\n      } else if (!isOptional || param !== undefined) {\n        if (!Array.isArray(param)) return res.sendStatus(400)\n\n        const vals = (param as string[]).map(Number)\n\n        if (vals.some(isNaN)) return res.sendStatus(400)\n\n        query[key] = vals as any\n      }\n    } else if (!isOptional || param !== undefined) {\n      const val = Number(param)\n\n      if (isNaN(val)) return res.sendStatus(400)\n\n      query[key] = val as any\n    }\n  }\n\n  next()\n}\n"
            : '') + (hasBooleanTypeQuery
            ? "\nconst parseBooleanTypeQueryParams = (booleanTypeParams: [string, boolean, boolean][]): RequestHandler => ({ query }, res, next) => {\n  for (const [key, isOptional, isArray] of booleanTypeParams) {\n    const param = query[key]\n\n    if (isArray) {\n      if (!isOptional && param === undefined) {\n        query[key] = []\n      } else if (!isOptional || param !== undefined) {\n        if (!Array.isArray(param)) return res.sendStatus(400)\n\n        const vals = (param as string[]).map(p => p === 'true' ? true : p === 'false' ? false : null)\n\n        if (vals.some(v => v === null)) return res.sendStatus(400)\n\n        query[key] = vals as any\n      }\n    } else if (!isOptional || param !== undefined) {\n      const val = param === 'true' ? true : param === 'false' ? false : null\n\n      if (val === null) return res.sendStatus(400)\n\n      query[key] = val as any\n    }\n  }\n\n  next()\n}\n"
            : '') + (hasOptionalQuery
            ? "\nconst callParserIfExistsQuery = (parser: RequestHandler): RequestHandler => (req, res, next) =>\n  Object.keys(req.query).length ? parser(req, res, next) : next()\n"
            : '') + (hasJSONBody
            ? "\nconst parseJSONBoby: RequestHandler = (req, res, next) => {\n  express.json()(req, res, err => {\n    if (err) return res.sendStatus(400)\n\n    next()\n  })\n}\n"
            : '') + (hasTypedParams
            ? "\nconst createTypedParamsHandler = (numberTypeParams: string[]): RequestHandler => (req, res, next) => {\n  const params: Record<string, string | number> = req.params\n\n  for (const key of numberTypeParams) {\n    const val = Number(params[key])\n\n    if (isNaN(val)) return res.sendStatus(400)\n\n    params[key] = val\n  }\n\n  next()\n}\n"
            : '') + (hasValidator
            ? "\nconst createValidateHandler = (validators: (req: Request) => (Promise<void> | null)[]): RequestHandler =>\n  (req, res, next) => Promise.all(validators(req)).then(() => next()).catch(err => res.status(400).send(err))\n"
            : '') + (hasMulter
            ? "\nconst formatMulterData = (arrayTypeKeys: [string, boolean][]): RequestHandler => ({ body, files }, _res, next) => {\n  for (const [key] of arrayTypeKeys) {\n    if (body[key] === undefined) body[key] = []\n    else if (!Array.isArray(body[key])) {\n      body[key] = [body[key]]\n    }\n  }\n\n  for (const file of files as MulterFile[]) {\n    if (Array.isArray(body[file.fieldname])) {\n      body[file.fieldname].push(file)\n    } else {\n      body[file.fieldname] = file\n    }\n  }\n\n  for (const [key, isOptional] of arrayTypeKeys) {\n    if (!body[key].length && isOptional) delete body[key]\n  }\n\n  next()\n}\n"
            : '') + (hasMethodToHandler ? genHandlerText(false) : '') + (hasAsyncMethodToHandler ? genHandlerText(true) : '') + (hasMethodToHandlerWithSchema ? genHandlerWithSchemaText(false) : '') + (hasAsyncMethodToHandlerWithSchema ? genHandlerWithSchemaText(true) : '') + "\nexport default (app: Express, options: FrourioOptions = {}) => {\n  const basePath = options.basePath ?? ''\n" + (hasValidator
            ? '  const transformerOptions: ClassTransformOptions = { enableCircularCheck: true, ...options.transformer }\n' +
                '  const validatorOptions: ValidatorOptions = { validationError: { target: false }, ...options.validator }\n'
            : '') + consts + (hasMulter
            ? "  const uploader = multer({ dest: path.join(__dirname, '.upload'), limits: { fileSize: 1024 ** 3 }, ...options.multer }).any()\n"
            : '') + "\n" + controllers + "\n  return app\n}\n",
        filePath: path_1.default.posix.join(input, '$server.ts')
    };
});
//# sourceMappingURL=buildServerFile.js.map