"use strict";
var __read = (this && this.__read) || function (o, n) {
    var m = typeof Symbol === "function" && o[Symbol.iterator];
    if (!m) return o;
    var i = m.call(o), r, ar = [], e;
    try {
        while ((n === void 0 || n-- > 0) && !(r = i.next()).done) ar.push(r.value);
    }
    catch (error) { e = { error: error }; }
    finally {
        try {
            if (r && !r.done && (m = i["return"])) m.call(i);
        }
        finally { if (e) throw e.error; }
    }
    return ar;
};
var __spreadArray = (this && this.__spreadArray) || function (to, from) {
    for (var i = 0, il = from.length, j = to.length; i < il; i++, j++)
        to[j] = from[i];
    return to;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
var path_1 = __importDefault(require("path"));
var fs_1 = __importDefault(require("fs"));
var typescript_1 = __importDefault(require("typescript"));
var createDefaultFilesIfNotExists_1 = __importDefault(require("./createDefaultFilesIfNotExists"));
var findRootFiles = function (dir) {
    return fs_1.default
        .readdirSync(dir, { withFileTypes: true })
        .reduce(function (prev, d) { return __spreadArray(__spreadArray([], __read(prev)), __read((d.isDirectory()
        ? findRootFiles(dir + "/" + d.name)
        : d.name === 'hooks.ts' || d.name === 'controller.ts'
            ? [dir + "/" + d.name]
            : []))); }, []);
};
var initTSC = function (appDir, project) {
    var configDir = path_1.default.resolve(project.replace(/\/[^/]+\.json$/, ''));
    var configFileName = typescript_1.default.findConfigFile(configDir, typescript_1.default.sys.fileExists, project.endsWith('.json') ? project.split('/').pop() : undefined);
    var compilerOptions = configFileName
        ? typescript_1.default.parseJsonConfigFileContent(typescript_1.default.readConfigFile(configFileName, typescript_1.default.sys.readFile).config, typescript_1.default.sys, configDir)
        : undefined;
    var program = typescript_1.default.createProgram(findRootFiles(appDir), (compilerOptions === null || compilerOptions === void 0 ? void 0 : compilerOptions.options)
        ? { baseUrl: compilerOptions === null || compilerOptions === void 0 ? void 0 : compilerOptions.options.baseUrl, paths: compilerOptions === null || compilerOptions === void 0 ? void 0 : compilerOptions.options.paths }
        : {});
    return { program: program, checker: program.getTypeChecker() };
};
var createRelayFile = function (input, appText, additionalReqs, params) {
    var hasAdditionals = !!additionalReqs.length;
    var hasMultiAdditionals = additionalReqs.length > 1;
    var text = "import type { Injectable } from 'velona'\nimport { depend } from 'velona'\nimport type { Express, RequestHandler } from 'express'\nimport type { Schema } from 'fast-json-stringify'\nimport type { HttpStatusOk } from 'aspida'\nimport type { ServerMethods } from '" + appText + "'\n" + (hasMultiAdditionals
        ? additionalReqs
            .map(function (req, i) {
            return "import type { AdditionalRequest as AdditionalRequest" + i + " } from '" + req.replace(/^\.\/\./, '.') + "'\n";
        })
            .join('')
        : hasAdditionals
            ? "import type { AdditionalRequest } from '" + additionalReqs[0] + "'\n"
            : '') + "import type { Methods } from './'\n\n" + (hasMultiAdditionals
        ? "type AdditionalRequest = " + additionalReqs
            .map(function (_, i) { return "AdditionalRequest" + i; })
            .join(' & ') + "\n"
        : '') + (hasAdditionals
        ? 'type AddedRequestHandler = RequestHandler extends (req: infer U, ...args: infer V) => infer W ? (req: U & Partial<AdditionalRequest>, ...args: V) => W : never\n'
        : '') + "type Hooks = {\n" + ['onRequest', 'preParsing', 'preValidation', 'preHandler']
        .map(function (key) {
        return hasAdditionals
            ? "  " + key + "?: AddedRequestHandler | AddedRequestHandler[]\n"
            : "  " + key + "?: RequestHandler | RequestHandler[]\n";
    })
        .join('') + "}\ntype ControllerMethods = ServerMethods<Methods, " + (hasAdditionals ? 'AdditionalRequest & ' : '') + "{" + (params.length
        ? "\n  params: {\n" + params.map(function (v) { return "    " + v[0] + ": " + v[1]; }).join('\n') + "\n  }\n"
        : '') + "}>\n\nexport function defineResponseSchema<T extends { [U in keyof ControllerMethods]?: { [V in HttpStatusOk]?: Schema }}>(methods: () => T) {\n  return methods\n}\n\nexport function defineHooks<T extends Hooks>(hooks: (app: Express) => T): (app: Express) => T\nexport function defineHooks<T extends Record<string, any>, U extends Hooks>(deps: T, cb: (d: T, app: Express) => U): Injectable<T, [Express], U>\nexport function defineHooks<T extends Record<string, any>>(hooks: (app: Express) => Hooks | T, cb?: (deps: T, app: Express) => Hooks) {\n  return cb && typeof hooks !== 'function' ? depend(hooks, cb) : hooks\n}\n\nexport function defineController(methods: (app: Express) => ControllerMethods): (app: Express) => ControllerMethods\nexport function defineController<T extends Record<string, any>>(deps: T, cb: (d: T, app: Express) => ControllerMethods): Injectable<T, [Express], ControllerMethods>\nexport function defineController<T extends Record<string, any>>(methods: (app: Express) => ControllerMethods | T, cb?: (deps: T, app: Express) => ControllerMethods) {\n  return cb && typeof methods !== 'function' ? depend(methods, cb) : methods\n}\n";
    fs_1.default.writeFileSync(path_1.default.join(input, '$relay.ts'), text.replace(', {}', '').replace(' & {}', ''), 'utf8');
};
var getAdditionalResPath = function (input, name) {
    return fs_1.default.existsSync(path_1.default.join(input, name + ".ts")) &&
        /(^|\n)export .+ AdditionalRequest(,| )/.test(fs_1.default.readFileSync(path_1.default.join(input, name + ".ts"), 'utf8'))
        ? ["./" + name]
        : [];
};
var createFiles = function (appDir, dirPath, params, appPath, additionalRequestPaths) {
    var input = path_1.default.posix.join(appDir, dirPath);
    var appText = "../" + appPath;
    var additionalReqs = __spreadArray(__spreadArray([], __read(additionalRequestPaths.map(function (p) { return "./." + p; }))), __read(getAdditionalResPath(input, 'hooks')));
    createDefaultFilesIfNotExists_1.default(input);
    createRelayFile(input, appText, __spreadArray(__spreadArray([], __read(additionalReqs)), __read(getAdditionalResPath(input, 'controller'))), params);
    fs_1.default.readdirSync(input, { withFileTypes: true }).forEach(function (d) {
        var _a;
        return d.isDirectory() &&
            createFiles(appDir, path_1.default.posix.join(dirPath, d.name), d.name.startsWith('_')
                ? __spreadArray(__spreadArray([], __read(params)), [[d.name.slice(1).split('@')[0], (_a = d.name.split('@')[1]) !== null && _a !== void 0 ? _a : 'string']]) : params, appText, additionalReqs);
    });
};
exports.default = (function (appDir, project) {
    createFiles(appDir, '', [], '$server', []);
    var _a = initTSC(appDir, project), program = _a.program, checker = _a.checker;
    var hooksPaths = [];
    var controllers = [];
    var createText = function (dirPath, cascadingHooks) {
        var input = path_1.default.posix.join(appDir, dirPath);
        var source = program.getSourceFile(path_1.default.join(input, 'index.ts'));
        var results = [];
        var hooks = cascadingHooks;
        if (source) {
            var methods = typescript_1.default.forEachChild(source, function (node) {
                var _a;
                return (typescript_1.default.isTypeAliasDeclaration(node) || typescript_1.default.isInterfaceDeclaration(node)) &&
                    node.name.escapedText === 'Methods' &&
                    ((_a = node.modifiers) === null || _a === void 0 ? void 0 : _a.some(function (m) { return m.kind === typescript_1.default.SyntaxKind.ExportKeyword; }))
                    ? checker.getTypeAtLocation(node).getProperties()
                    : undefined;
            });
            var hooksSource = program.getSourceFile(path_1.default.join(input, 'hooks.ts'));
            if (hooksSource) {
                var events = typescript_1.default.forEachChild(hooksSource, function (node) {
                    if (typescript_1.default.isExportAssignment(node)) {
                        return node.forEachChild(function (node) {
                            return typescript_1.default.isCallExpression(node) &&
                                node.forEachChild(function (node) {
                                    if (typescript_1.default.isMethodDeclaration(node) ||
                                        typescript_1.default.isArrowFunction(node) ||
                                        typescript_1.default.isFunctionDeclaration(node)) {
                                        return (node.body &&
                                            checker
                                                .getTypeAtLocation(node.body)
                                                .getProperties()
                                                .map(function (p) {
                                                var typeNode = p.valueDeclaration &&
                                                    checker.typeToTypeNode(checker.getTypeOfSymbolAtLocation(p, p.valueDeclaration), undefined, undefined);
                                                return {
                                                    type: p.name,
                                                    isArray: typeNode
                                                        ? typescript_1.default.isArrayTypeNode(typeNode) || typescript_1.default.isTupleTypeNode(typeNode)
                                                        : false
                                                };
                                            }));
                                    }
                                });
                        });
                    }
                });
                if (events) {
                    hooks = __spreadArray(__spreadArray([], __read(cascadingHooks)), [{ name: "hooks" + hooksPaths.length, events: events }]);
                    hooksPaths.push(input + "/hooks");
                }
            }
            if (methods === null || methods === void 0 ? void 0 : methods.length) {
                var controllerSource = program.getSourceFile(path_1.default.join(input, 'controller.ts'));
                var isPromiseMethods_1 = [];
                var ctrlHooksSignature = void 0;
                var resSchemaSignature = void 0;
                if (controllerSource) {
                    isPromiseMethods_1 =
                        typescript_1.default.forEachChild(controllerSource, function (node) {
                            return typescript_1.default.isExportAssignment(node) &&
                                node.forEachChild(function (nod) {
                                    return typescript_1.default.isCallExpression(nod) &&
                                        checker
                                            .getSignaturesOfType(checker.getTypeAtLocation(nod.arguments[nod.arguments.length - 1]), typescript_1.default.SignatureKind.Call)[0]
                                            .getReturnType()
                                            .getProperties()
                                            .map(function (t) {
                                            var _a;
                                            return t.valueDeclaration &&
                                                ((_a = checker
                                                    .getSignaturesOfType(checker.getTypeOfSymbolAtLocation(t, t.valueDeclaration), typescript_1.default.SignatureKind.Call)[0]
                                                    .getReturnType()
                                                    .getSymbol()) === null || _a === void 0 ? void 0 : _a.getEscapedName()) === 'Promise' &&
                                                t.name;
                                        })
                                            .filter(function (n) { return !!n; });
                                });
                        }) || [];
                    var ctrlHooksNode_1;
                    var resSchemaNode_1;
                    typescript_1.default.forEachChild(controllerSource, function (node) {
                        var _a, _b, _c, _d, _e;
                        if (typescript_1.default.isVariableStatement(node) &&
                            ((_a = node.modifiers) === null || _a === void 0 ? void 0 : _a.some(function (m) { return m.kind === typescript_1.default.SyntaxKind.ExportKeyword; }))) {
                            ctrlHooksNode_1 =
                                (_b = node.declarationList.declarations.find(function (d) { return d.name.getText() === 'hooks'; })) !== null && _b !== void 0 ? _b : ctrlHooksNode_1;
                            resSchemaNode_1 =
                                (_c = node.declarationList.declarations.find(function (d) { return d.name.getText() === 'responseSchema'; })) !== null && _c !== void 0 ? _c : resSchemaNode_1;
                        }
                        else if (typescript_1.default.isExportDeclaration(node)) {
                            var exportClause = node.exportClause;
                            if (exportClause && typescript_1.default.isNamedExports(exportClause)) {
                                ctrlHooksNode_1 =
                                    (_d = exportClause.elements.find(function (el) { return el.name.text === 'hooks'; })) !== null && _d !== void 0 ? _d : ctrlHooksNode_1;
                                resSchemaNode_1 =
                                    (_e = exportClause.elements.find(function (el) { return el.name.text === 'responseSchema'; })) !== null && _e !== void 0 ? _e : resSchemaNode_1;
                            }
                        }
                    });
                    if (ctrlHooksNode_1) {
                        ctrlHooksSignature = checker.getSignaturesOfType(checker.getTypeAtLocation(ctrlHooksNode_1), typescript_1.default.SignatureKind.Call)[0];
                    }
                    if (resSchemaNode_1) {
                        resSchemaSignature = checker.getSignaturesOfType(checker.getTypeAtLocation(resSchemaNode_1), typescript_1.default.SignatureKind.Call)[0];
                    }
                }
                var ctrlHooksEvents_1 = ctrlHooksSignature === null || ctrlHooksSignature === void 0 ? void 0 : ctrlHooksSignature.getReturnType().getProperties().map(function (p) {
                    var typeNode = p.valueDeclaration &&
                        checker.typeToTypeNode(checker.getTypeOfSymbolAtLocation(p, p.valueDeclaration), undefined, undefined);
                    return {
                        type: p.name,
                        isArray: typeNode
                            ? typescript_1.default.isArrayTypeNode(typeNode) || typescript_1.default.isTupleTypeNode(typeNode)
                            : false
                    };
                });
                var genHookTexts_1 = function (event) {
                    var _a;
                    return __spreadArray(__spreadArray([], __read(hooks.reduce(function (prev, h) {
                        var ev = h.events.find(function (e) { return e.type === event; });
                        return ev ? __spreadArray(__spreadArray([], __read(prev)), ["" + (ev.isArray ? '...' : '') + h.name + "." + event]) : prev;
                    }, []))), __read(((_a = ctrlHooksEvents_1 === null || ctrlHooksEvents_1 === void 0 ? void 0 : ctrlHooksEvents_1.map(function (e) {
                        return e.type === event
                            ? (e.isArray ? '...' : '') + "ctrlHooks" + controllers.filter(function (c) { return c[1]; }).length + "." + event
                            : '';
                    })) !== null && _a !== void 0 ? _a : [])));
                };
                var resSchemaMethods_1 = resSchemaSignature === null || resSchemaSignature === void 0 ? void 0 : resSchemaSignature.getReturnType().getProperties().map(function (p) { return p.name; });
                var genResSchemaText_1 = function (method) {
                    return "responseSchema" + controllers.filter(function (c) { return c[2]; }).length + "." + method;
                };
                var getSomeTypeQueryParams_1 = function (typeName, query) {
                    return query.valueDeclaration &&
                        checker
                            .getTypeOfSymbolAtLocation(query, query.valueDeclaration)
                            .getProperties()
                            .map(function (p) {
                            var _a;
                            var typeString = p.valueDeclaration &&
                                checker.typeToString(checker.getTypeOfSymbolAtLocation(p, p.valueDeclaration));
                            return typeString === typeName || typeString === typeName + "[]"
                                ? "['" + p.name + "', " + !!((_a = p.declarations) === null || _a === void 0 ? void 0 : _a.some(function (d) {
                                    return d.getChildren().some(function (c) { return c.kind === typescript_1.default.SyntaxKind.QuestionToken; });
                                })) + ", " + (typeString === typeName + "[]") + "]"
                                : null;
                        })
                            .filter(Boolean);
                };
                results.push(methods
                    .map(function (m) {
                    var _a, _b;
                    var props = m.valueDeclaration
                        ? checker.getTypeOfSymbolAtLocation(m, m.valueDeclaration).getProperties()
                        : [];
                    var query = props.find(function (p) { return p.name === 'query'; });
                    var numberTypeQueryParams = query && getSomeTypeQueryParams_1('number', query);
                    var booleanTypeQueryParams = query && getSomeTypeQueryParams_1('boolean', query);
                    var validateInfo = [
                        { name: 'query', val: query },
                        { name: 'body', val: props.find(function (p) { return p.name === 'reqBody'; }) },
                        { name: 'headers', val: props.find(function (p) { return p.name === 'reqHeaders'; }) }
                    ]
                        .filter(function (prop) { return !!prop.val; })
                        .map(function (_a) {
                        var _b;
                        var name = _a.name, val = _a.val;
                        return ({
                            name: name,
                            type: val.valueDeclaration &&
                                checker.getTypeOfSymbolAtLocation(val, val.valueDeclaration),
                            hasQuestion: !!((_b = val.declarations) === null || _b === void 0 ? void 0 : _b.some(function (d) { return d.getChildAt(1).kind === typescript_1.default.SyntaxKind.QuestionToken; }))
                        });
                    })
                        .filter(function (_a) {
                        var type = _a.type;
                        return type === null || type === void 0 ? void 0 : type.isClass();
                    });
                    var reqFormat = props.find(function (p) { return p.name === 'reqFormat'; });
                    var isFormData = ((reqFormat === null || reqFormat === void 0 ? void 0 : reqFormat.valueDeclaration) &&
                        checker.typeToString(checker.getTypeOfSymbolAtLocation(reqFormat, reqFormat.valueDeclaration))) === 'FormData';
                    var reqBody = props.find(function (p) { return p.name === 'reqBody'; });
                    var handlers = __spreadArray(__spreadArray(__spreadArray(__spreadArray(__spreadArray(__spreadArray(__spreadArray(__spreadArray(__spreadArray([], __read(genHookTexts_1('onRequest'))), __read(genHookTexts_1('preParsing'))), [
                        (numberTypeQueryParams === null || numberTypeQueryParams === void 0 ? void 0 : numberTypeQueryParams.length)
                            ? ((_a = query === null || query === void 0 ? void 0 : query.declarations) === null || _a === void 0 ? void 0 : _a.some(function (d) { return d.getChildAt(1).kind === typescript_1.default.SyntaxKind.QuestionToken; }))
                                ? "callParserIfExistsQuery(parseNumberTypeQueryParams([" + numberTypeQueryParams.join(', ') + "]))"
                                : "parseNumberTypeQueryParams([" + numberTypeQueryParams.join(', ') + "])"
                            : '',
                        (booleanTypeQueryParams === null || booleanTypeQueryParams === void 0 ? void 0 : booleanTypeQueryParams.length)
                            ? ((_b = query === null || query === void 0 ? void 0 : query.declarations) === null || _b === void 0 ? void 0 : _b.some(function (d) { return d.getChildAt(1).kind === typescript_1.default.SyntaxKind.QuestionToken; }))
                                ? "callParserIfExistsQuery(parseBooleanTypeQueryParams([" + booleanTypeQueryParams.join(', ') + "]))"
                                : "parseBooleanTypeQueryParams([" + booleanTypeQueryParams.join(', ') + "])"
                            : ''
                    ]), __read((isFormData && (reqBody === null || reqBody === void 0 ? void 0 : reqBody.valueDeclaration)
                        ? [
                            'uploader',
                            "formatMulterData([" + checker
                                .getTypeOfSymbolAtLocation(reqBody, reqBody.valueDeclaration)
                                .getProperties()
                                .map(function (p) {
                                var _a;
                                var node = p.valueDeclaration &&
                                    checker.typeToTypeNode(checker.getTypeOfSymbolAtLocation(p, p.valueDeclaration), undefined, undefined);
                                return node && (typescript_1.default.isArrayTypeNode(node) || typescript_1.default.isTupleTypeNode(node))
                                    ? "['" + p.name + "', " + !!((_a = p.declarations) === null || _a === void 0 ? void 0 : _a.some(function (d) {
                                        return d.getChildren().some(function (c) { return c.kind === typescript_1.default.SyntaxKind.QuestionToken; });
                                    })) + "]"
                                    : undefined;
                            })
                                .filter(Boolean)
                                .join(', ') + "])"
                        ]
                        : []))), [
                        !reqFormat && reqBody ? 'parseJSONBoby' : ''
                    ]), __read(genHookTexts_1('preValidation'))), [
                        validateInfo.length
                            ? "createValidateHandler(req => [\n" + validateInfo
                                .map(function (v) {
                                return v.type
                                    ? "      " + (v.hasQuestion ? "Object.keys(req." + v.name + ").length ? " : '') + "validateOrReject(plainToInstance(Validators." + checker.typeToString(v.type) + ", req." + v.name + ", transformerOptions), validatorOptions)" + (v.hasQuestion ? ' : null' : '')
                                    : '';
                            })
                                .join(',\n') + "\n    ])"
                            : '',
                        dirPath.includes('@number')
                            ? "createTypedParamsHandler(['" + dirPath
                                .split('/')
                                .filter(function (p) { return p.includes('@number'); })
                                .map(function (p) { return p.split('@')[0].slice(1); })
                                .join("', '") + "'])"
                            : ''
                    ]), __read(genHookTexts_1('preHandler'))), [
                        (resSchemaMethods_1 === null || resSchemaMethods_1 === void 0 ? void 0 : resSchemaMethods_1.includes(m.name))
                            ? (isPromiseMethods_1.includes(m.name)
                                ? 'asyncMethodToHandlerWithSchema'
                                : 'methodToHandlerWithSchema') + "(controller" + controllers.length + "." + m.name + ", " + genResSchemaText_1(m.name) + ")"
                            : (isPromiseMethods_1.includes(m.name) ? 'asyncMethodToHandler' : 'methodToHandler') + "(controller" + controllers.length + "." + m.name + ")"
                    ]).filter(Boolean);
                    return "  app." + m.name + "(`${basePath}" + ("/" + dirPath)
                        .replace(/\/_/g, '/:')
                        .replace(/@.+?($|\/)/g, '$1') + "`, " + (handlers.length === 1 ? handlers[0] : "[\n    " + handlers.join(',\n    ') + "\n  ]") + ")\n";
                })
                    .join('\n'));
                controllers.push([input + "/controller", !!ctrlHooksEvents_1, !!resSchemaMethods_1]);
            }
        }
        var childrenDirs = fs_1.default.readdirSync(input, { withFileTypes: true }).filter(function (d) { return d.isDirectory(); });
        if (childrenDirs.length) {
            results.push.apply(results, __spreadArray([], __read(childrenDirs
                .filter(function (d) { return !d.name.startsWith('_'); })
                .reduce(function (prev, d) { return __spreadArray(__spreadArray([], __read(prev)), __read(createText(path_1.default.posix.join(dirPath, d.name), hooks))); }, []))));
            var value = childrenDirs.find(function (d) { return d.name.startsWith('_'); });
            if (value) {
                results.push.apply(results, __spreadArray([], __read(createText(path_1.default.posix.join(dirPath, value.name), hooks))));
            }
        }
        return results;
    };
    var text = createText('', []).join('\n');
    var ctrlHooks = controllers.filter(function (c) { return c[1]; });
    var resSchemas = controllers.filter(function (c) { return c[2]; });
    return {
        imports: "" + hooksPaths
            .map(function (m, i) {
            return "import hooksFn" + i + " from '" + m.replace(/^api/, './api').replace(appDir, './api') + "'\n";
        })
            .join('') + controllers
            .map(function (ctrl, i) {
            return "import controllerFn" + i + (ctrl[1] || ctrl[2]
                ? ", { " + (ctrl[1] ? "hooks as ctrlHooksFn" + ctrlHooks.indexOf(ctrl) : '') + (ctrl[1] && ctrl[2] ? ', ' : '') + (ctrl[2] ? "responseSchema as responseSchemaFn" + resSchemas.indexOf(ctrl) : '') + " }"
                : '') + " from '" + ctrl[0].replace(/^api/, './api').replace(appDir, './api') + "'\n";
        })
            .join(''),
        consts: "" + hooksPaths
            .map(function (_, i) { return "  const hooks" + i + " = hooksFn" + i + "(app)\n"; })
            .join('') + ctrlHooks
            .map(function (_, i) { return "  const ctrlHooks" + i + " = ctrlHooksFn" + i + "(app)\n"; })
            .join('') + resSchemas
            .map(function (_, i) { return "  const responseSchema" + i + " = responseSchemaFn" + i + "()\n"; })
            .join('') + controllers
            .map(function (_, i) { return "  const controller" + i + " = controllerFn" + i + "(app)\n"; })
            .join(''),
        controllers: text
    };
});
//# sourceMappingURL=createControllersText.js.map